#include "contiki.h"
#include "net/rime.h"
#include "net/rime/rimebuf.h"
#include "dev/button-sensor.h"
#include "dev/leds.h"
#include <stdio.h>
#include "dev/cc2420.h"
#include "net/rime/route.h"
#include "route_node.h"

struct SLAVE_INFO
{
  rimeaddr_t slave_addr[MAXSLAVE];
  struct route_entry rt[MAXSLAVE][MAX_RT];
  int rt_count[MAXSLAVE];
  int is_active[MAXSLAVE];
  int test_flag[MAXSLAVE];
  int is_fault_timer_set[MAXSLAVE];
  struct etimer fault_timer[MAXSLAVE];
  int recv_time[MAXSLAVE][3];
  int diff_time[MAXSLAVE];
  int count;
};

struct SLAVE_INFO slaveinfo;

static char coord_found = 0;
static int coord_addr[2];
static int my_hop_count = 0;
/*---------------------------------------------------------------------------*/
PROCESS(example_unicast_process, "coord router");
PROCESS(coord_process2, "coord_process2");
AUTOSTART_PROCESSES(&example_unicast_process, &coord_process2);
/*---------------------------------------------------------------------------*/
static void
recv_from_broadcast(struct broadcast_conn *broadcast, rimeaddr_t *from)
{
    printf("COORD ROUTER hears bcast from %d.%d\n", from->u8[0] , from->u8[1] );
}
static const struct broadcast_callbacks bc = {recv_from_broadcast};
static struct broadcast_conn broadcast_conn;

/*---------------------------------------------------------------------------*/
static void recv_runicast(struct runicast_conn *c, rimeaddr_t *from, uint8_t seqno)
{
  printf("COORD ROUTER UNICAST RECEIVED FROM:%d.%d\n",  from->u8[0], from->u8[1]);
  leds_invert( LEDS_RED );
  struct MSG recmsg = *( struct MSG* )rimebuf_dataptr();
  int i = 0, j=0, done = 0;
  if( recmsg.msg_type == ROUTERTABLE ) {
    if(slaveinfo.count < MAXSLAVE) {
      //If already existing then update
      
      for(i=0;i<slaveinfo.count;i++) {
        if(rimeaddr_cmp(&slaveinfo.slave_addr[i], &recmsg.from_addr)) {
          printf("COORD ROUTER: Updating slave info. Slave match found.\n");
          slaveinfo.rt_count[i] = recmsg.rt_count;
          for(j=0;j<recmsg.rt_count;j++) {
            memcpy(&slaveinfo.rt[i][j], &recmsg.rt[j], sizeof(struct route_entry));
          }
          done = 1;
        }
      }
      //Else add new entry
      if(!done) {
        printf("COORD ROUTER: Adding new slave info of %d.%d.\n",recmsg.from_addr.u8[0], recmsg.from_addr.u8[1]);
        rimeaddr_copy(&slaveinfo.slave_addr[slaveinfo.count], &recmsg.from_addr);
        slaveinfo.rt_count[slaveinfo.count] = recmsg.rt_count;
        for(j=0;j<recmsg.rt_count;j++) {
          memcpy(&slaveinfo.rt[slaveinfo.count][j], &recmsg.rt[j], sizeof(struct route_entry));
        }
        slaveinfo.diff_time[slaveinfo.count] = 0;
        slaveinfo.recv_time[slaveinfo.count][0] = 0;
        slaveinfo.recv_time[slaveinfo.count][1] = 0;
        slaveinfo.recv_time[slaveinfo.count][2] = 0;
        slaveinfo.is_active[slaveinfo.count] = 1;
        slaveinfo.is_fault_timer_set[slaveinfo.count] = 0;
        slaveinfo.count++;
      }
    }
  }
  else if( recmsg.msg_type == DATA ) {
    for(i=0;i<slaveinfo.count;i++) {
      if(rimeaddr_cmp(&slaveinfo.slave_addr[i], &recmsg.from_addr)) {
        printf("COORD ROUTER: Processing data from %d.%d.\n", recmsg.from_addr.u8[0], recmsg.from_addr.u8[1]);
        slaveinfo.test_flag[i] = 1;
	slaveinfo.is_active[i] = 1;
        if(slaveinfo.diff_time[i] <= 0) {
          for(j=0;j<3;j++) {
            if(slaveinfo.recv_time[i][j] == 0) {
              slaveinfo.recv_time[i][j] = clock_time()/CLOCK_SECOND;
              break; //break from loop j
            }
          }
          slaveinfo.diff_time[i] = slaveinfo.recv_time[i][2] - slaveinfo.recv_time[i][0];
          /*if(slaveinfo.diff_time[i] > 0) {
            etimer_set(&slaveinfo.fault_timer[i], slaveinfo.diff_time[i]);
            slaveinfo.test_flag[i] = 0;
            printf("COORD ROUTER: Processing data from %d.%d. Setting timer %d\n", recmsg.from_addr.u8[0], recmsg.from_addr.u8[1], slaveinfo.diff_time[i]);
          }*/
        }//end if(slaveinfo.diff_time[i] <= 0)
        break; //break from loop i
      } //end if(rimeaddr_cmp...
    } //end for(i=0...
    
  } //end else if

}

static void sent_runicast(struct runicast_conn *c, rimeaddr_t *to, uint8_t retransmissions)
{
  //printf("Sink: runicast message sent to %d.%d, retransmissions %d\n", to->u8[0], to->u8[1], retransmissions);
}
static void
timedout_runicast(struct runicast_conn *c, rimeaddr_t *to, uint8_t retransmissions)
{
  printf("ROUTER: runicast message timed out when sending to %d.%d, retransmissions %d\n", to->u8[0], to->u8[1], retransmissions);
}
static const struct runicast_callbacks runicast_callbacks = {recv_runicast,
						   sent_runicast,
						   timedout_runicast};
static struct runicast_conn runicast;

/*--------------------------------------------------------------------------*/
PROCESS_THREAD(coord_process2, ev, data)
{
  PROCESS_BEGIN();
  static struct etimer p2_t;
  etimer_set(&p2_t, CLOCK_SECOND*20);
  while(1) {
  printf("COORD: p2: Before wait event\n");
  PROCESS_WAIT_EVENT();
  printf("COORD: p2: Some event\n");
  if(ev == PROCESS_EVENT_TIMER) {
    int i = 0;
    int j=0;
    if(etimer_expired(&p2_t)) {
      printf("COORD: p2: woke up due to p2_t\n");
      for(i=0;i<slaveinfo.count;i++) {
        if(slaveinfo.diff_time[i] > 0) {
          if (slaveinfo.is_fault_timer_set[i] == 0) {
            etimer_set(&slaveinfo.fault_timer[i], slaveinfo.diff_time[i]*CLOCK_SECOND);
            slaveinfo.test_flag[i] = 0;
            slaveinfo.is_fault_timer_set[i] = 1;
            printf("COORD: p2: Processing data from %d.%d. Setting timer %d\n", 
                 slaveinfo.slave_addr[i].u8[0], slaveinfo.slave_addr[i].u8[1], slaveinfo.diff_time[i]);
          }
          else
            j++;
        }
      }
      if(j < slaveinfo.count) 
        etimer_set(&p2_t, CLOCK_SECOND*20);
      else
        etimer_set(&p2_t, CLOCK_SECOND*60);
      continue;
    }
    for(i=0;i<slaveinfo.count;i++) {
      if(slaveinfo.is_active[i] == 1 && slaveinfo.diff_time[i] > 0 && etimer_expired(&slaveinfo.fault_timer[i])) {
        printf("COORD: p2: timer expired for %d.%d timer %p\n",slaveinfo.slave_addr[i].u8[0],slaveinfo.slave_addr[i].u8[1], &slaveinfo.fault_timer[i]);
        if(slaveinfo.test_flag[i] == 0) {
          printf("COORD: node %d.%d failed.\n", slaveinfo.slave_addr[i].u8[0],slaveinfo.slave_addr[i].u8[1]);
/*DONT STOP THE TIMER, KEEP IT RUNNING, SO THAT IF THE NODE RECOVERS LATER, NO ISSUES
          etimer_stop(&slaveinfo.fault_timer[i]);
*/
          slaveinfo.is_active[i] = 0;
          continue;
        }
        slaveinfo.test_flag[i] = 0;
        etimer_set(&slaveinfo.fault_timer[i], slaveinfo.diff_time[i]*CLOCK_SECOND);
      }
    }
  }
  }
  PROCESS_END();
}

PROCESS_THREAD(example_unicast_process, ev, data)
{
  PROCESS_EXITHANDLER(runicast_close(&runicast);broadcast_close(&broadcast_conn);)
    
  PROCESS_BEGIN();
  slaveinfo.count = 0; //Init to zero
  cc2420_set_txpower( 12 );
  route_init(); //initialize routing table
  static struct etimer et;    
  etimer_set(&et, CLOCK_SECOND*2);
  PROCESS_WAIT_EVENT_UNTIL(etimer_expired(&et));

  rimeaddr_t addr , myaddr;

  /*******BROADCAST INITSINK MSG*****************/
  //delay so that all other motes are ON, need to do something later

  broadcast_open(&broadcast_conn , BCAST_CHANNEL , &bc);

  runicast_open(&runicast, UCAST_CHANNEL , &runicast_callbacks);

  while(1) 
  {
    PROCESS_WAIT_EVENT_UNTIL(ev == sensors_event && data == &button_sensor);

    struct MSG sinkmsg;
    sinkmsg.msg_type = ROUTERADVT;
    sinkmsg.data[0] = rimeaddr_node_addr.u8[0];
    sinkmsg.data[1] = rimeaddr_node_addr.u8[1];
    sinkmsg.data[2] = my_hop_count;
    sinkmsg.data[3] = rimeaddr_node_addr.u8[0];
    sinkmsg.data[4] = rimeaddr_node_addr.u8[0];
    rimebuf_copyfrom( &sinkmsg  , sizeof( struct MSG ));

    broadcast_send( &broadcast_conn );
    printf("coord router sends broadcast msg: msg size %d\n", sizeof(sinkmsg));
  }

  PROCESS_END();
}
/*---------------------------------------------------------------------------*/

