#include "contiki.h"
#include "net/rime.h"
#include "net/rime/rimebuf.h"
#include "dev/button-sensor.h"
#include "dev/leds.h"
#include <stdio.h>
#include "dev/cc2420.h"
#include "net/rime/route.h"
#include "lib/list.h"
#include "lib/memb.h"
#include "timer.h"
#include "sys/mt.h"
#include "random.h"
#include "route_node.h"

#define TXFAILURETHRESH 3

struct SLAVE_INFO
{
  rimeaddr_t slave_addr[MAXSLAVE];
  int count;
};

struct SLAVE_INFO slaveinfo;

static char coord_found = 0;
static int coord_addr[2];
static int my_hop_count = 9999;
static struct runicast_conn runicast;
static struct broadcast_conn broadcast_conn;
static char is_rt_sent = 0;
static char is_rt_updated = 0;
static struct mt_thread advt_thread;

//MEMB( received_bcasts , struct MSG , 8 );
/*---------------------------------------------------------------------------*/
PROCESS(example_unicast_process, "sink #1");
PROCESS(process2, "process2");
AUTOSTART_PROCESSES(&example_unicast_process, &process2);
/*---------------------------------------------------------------------------*/

static void recv_from_broadcast(struct broadcast_conn *broadcast, rimeaddr_t *from)
{    
    struct MSG recmsg = *( struct MSG* )rimebuf_dataptr();

	if( recmsg.msg_type == ROUTERADVT )
        {
          printf("Clock time is %d\n",clock_time()/CLOCK_SECOND);
	  if( recmsg.data[2] < my_hop_count )//the advertising router is at least at my level or towards coord
	  {
		  
		  //coord_found = 0;
	          rimeaddr_t dest , from;
        	  dest.u8[0] = recmsg.data[3];//this is the coordinatorś rime address
        	  dest.u8[1] = recmsg.data[4];
          
		  //take this as most current coord address
		  coord_addr[0] = recmsg.data[3];
		  coord_addr[1] = recmsg.data[4];

		  //update my hop count based on this
		  my_hop_count = recmsg.data[2] + 1;

		  from.u8[0] = recmsg.data[0];//the next hop address
 		  from.u8[1] = recmsg.data[1];
        	  if(!route_add( &dest , &from , recmsg.data[2] , 0 )) //dont care about time fields	
                  {
                    printf("ROUTER: updating routing table\n");
                    is_rt_updated = 1;
                  }
		  coord_found = 1;
           }
        }
    printf("ROUTER hears bcast from %d.%d. Total route %d\n", from->u8[0] , from->u8[1], route_num());    
}
static const struct broadcast_callbacks bc = {recv_from_broadcast};

/*---------------------------------------------------------------------------*/
static void
recv_runicast(struct runicast_conn *c, rimeaddr_t *from, uint8_t seqno)
{
  printf("ROUTER UNICAST RECEIVED FROM:%d.%d\n",  from->u8[0], from->u8[1]);
  leds_invert( LEDS_RED );
  /*****PROCESS THE DATA PACKET HERE************/
  struct MSG recmsg = *( struct MSG* )rimebuf_dataptr();
 
  if( recmsg.msg_type == DATA || recmsg.msg_type == ROUTERTABLE)
  {
	struct route_entry *t = route_get( 0 ) ; //get the head entry

	if( t != NULL )
	{
		rimeaddr_t nexthop = t->nexthop;
                recmsg.is_routed = 1;
                rimebuf_copyfrom( &recmsg  , sizeof( struct MSG ));
	        runicast_send( &runicast , &nexthop , MAX_RETRANSMISSIONS);
	}
	else
	{
		printf("ROUTER: routing Table Empty, not forwarding the packet\n");
	}
  }
}

static void sent_runicast(struct runicast_conn *c, rimeaddr_t *to, uint8_t retransmissions)
{
  //printf("Sink: runicast message sent to %d.%d, retransmissions %d\n", to->u8[0], to->u8[1], retransmissions);
}
static void timedout_runicast(struct runicast_conn *c, rimeaddr_t *to, uint8_t retransmissions)
{
  printf("ROUTER: runicast message timed out when sending to %d.%d, retransmissions %d\n", to->u8[0], to->u8[1], retransmissions);

  inc_tx_fail_count( 0 );

  if( get_tx_fail_count(0) > TXFAILURETHRESH )
  {	
	//remove that entry from routing table
	route_remove( route_get(0) );
	printf("ROUTER: next hop %d.%d seems to be dead, removed from routing table",to->u8[0], to->u8[1]);
	
  }

}

static const struct runicast_callbacks runicast_callbacks = {recv_runicast,
						   sent_runicast,
						   timedout_runicast};

/*--------------------------------------------------------------------------*/
PROCESS_THREAD(process2, ev, data)
{
  PROCESS_EXITHANDLER(broadcast_close(&broadcast_conn);)
  PROCESS_BEGIN();
  broadcast_open(&broadcast_conn , BCAST_CHANNEL , &bc);
  while(1) {
    static struct etimer p2_t;
    etimer_set(&p2_t, CLOCK_SECOND*40);
    PROCESS_WAIT_EVENT_UNTIL(etimer_expired(&p2_t));
	if(coord_found == 1 )
	{
		static struct etimer t;  
		etimer_set(&t,   10*(rimeaddr_node_addr.u8[0] + rimeaddr_node_addr.u8[1]) );	
		PROCESS_WAIT_EVENT_UNTIL(etimer_expired(&t));
                printf("P2: Router: test test bcast.\n");
		  struct MSG sinkmsg;
		  sinkmsg.msg_type = ROUTERADVT;
		  sinkmsg.data[3] = coord_addr[0];
		  sinkmsg.data[4] = coord_addr[1];
		  sinkmsg.data[0] = rimeaddr_node_addr.u8[0];//my rime address
		  sinkmsg.data[1] = rimeaddr_node_addr.u8[1];
		  sinkmsg.data[2] = my_hop_count;

		  rimebuf_copyfrom( &sinkmsg  , sizeof( struct MSG ));

		  broadcast_send( &broadcast_conn );
		  //printf("router advertises itself\n");

	}

  }
  PROCESS_END();
}

PROCESS_THREAD(example_unicast_process, ev, data)
{
  PROCESS_EXITHANDLER(runicast_close(&runicast);)//broadcast_close(&broadcast_conn);)
    
  PROCESS_BEGIN();

  cc2420_set_txpower( 15 );
  route_init(); //initialize routing table

  static struct etimer et;    
  etimer_set(&et, CLOCK_SECOND*1);
  PROCESS_WAIT_EVENT_UNTIL(etimer_expired(&et));

  rimeaddr_t addr , myaddr;

  //broadcast_open(&broadcast_conn , BCAST_CHANNEL , &bc);

  runicast_open(&runicast, UCAST_CHANNEL , &runicast_callbacks);
  while(1) 
  {
    //static struct etimer et;  
    static struct etimer et2;    
    //etimer_set(&et, CLOCK_SECOND*20);
    etimer_set(&et2, CLOCK_SECOND*23);
    PROCESS_WAIT_EVENT();
    if( etimer_expired(&et2) && coord_found)//ev == sensors_event && data == &button_sensor )
    {
	static struct etimer t1;  
	etimer_set(&t1,   10*(rimeaddr_node_addr.u8[0] + rimeaddr_node_addr.u8[1]) );
	PROCESS_WAIT_EVENT_UNTIL(etimer_expired(&t1));
     	struct route_entry *t = route_get( 0 ) ; //get the head entry

	if( t != NULL )
  	{
          rimeaddr_t nexthop = t->nexthop;
	  struct MSG sendmsg;
          if(is_rt_sent == 0 || is_rt_updated == 1)
          {
            printf("ROUTER:Sending routing table.\n");
	    sendmsg.msg_type = ROUTERTABLE;
            int i = 0;
            int num = route_num();
            if(num > MAX_RT) {
              num = MAX_RT;
            }
            sendmsg.rt_count = num;
            rimeaddr_copy(&sendmsg.from_addr, &rimeaddr_node_addr);
            for(i=0;i<num;i++) {
              memcpy(&sendmsg.rt[i],route_get(i), sizeof(struct route_entry));
            }
            is_rt_sent = 1;
            is_rt_updated = 0;
          }
          else
          {
             sendmsg.msg_type = DATA;
             rimeaddr_copy(&sendmsg.from_addr, &rimeaddr_node_addr);
          }
	  rimebuf_copyfrom( &sendmsg  , sizeof( struct MSG ));         
	  runicast_send(&runicast , &nexthop , MAX_RETRANSMISSIONS);	 
	  printf("ROUTER: Test Data unicast to: %d.%d\n", nexthop.u8[0] , nexthop.u8[1] );
  	}
  	else
  	{
	  printf("ROUTER: routing Table Empty\n");
  	}
        
        continue;   
     }
	/*if(etimer_expired(&et) && coord_found == 1 )
	{
		static struct etimer t;  
		etimer_set(&t,   10*(rimeaddr_node_addr.u8[0] + rimeaddr_node_addr.u8[1]) );
		PROCESS_WAIT_EVENT_UNTIL(etimer_expired(&t));
		  struct MSG sinkmsg;
		  sinkmsg.msg_type = ROUTERADVT;
		  sinkmsg.data[3] = coord_addr[0];
		  sinkmsg.data[4] = coord_addr[1];
		  sinkmsg.data[0] = rimeaddr_node_addr.u8[0];//my rime address
		  sinkmsg.data[1] = rimeaddr_node_addr.u8[1];
		  sinkmsg.data[2] = my_hop_count;

		  rimebuf_copyfrom( &sinkmsg  , sizeof( struct MSG ));

		  broadcast_send( &broadcast_conn );
		  //printf("router advertises itself\n");

	}*/

		  	
  }	
  PROCESS_END();
}
/*---------------------------------------------------------------------------*/

