#include<config.h>

#if defined(CONF_DSENSOR) && defined(CONF_DMOTOR)

#include<conio.h>
#include<unistd.h>
#include<time.h>
#include<lnp.h>
#include<lnp-logical.h>
#include<dsensor.h>
#include<dmotor.h>
#include<dkey.h>
#include<myrover.h>

#define LIGHT_SENSOR 	SENSOR_2
#define ROTATION_SENSOR SENSOR_3
#define LIGHT_READ	LIGHT_2

#define NEAR 0
#define FAR  1

// diretions reversed due to connection orientation
#define FWD rev
#define REV fwd

#define MOT_A 0
#define MOT_C 1
#define ALL   2

#define LEFT  0
#define RIGHT 1
#define FRWD  2
#define REVS  3

#define SPEED MAX_SPEED/3

#define GROUND_FOUND   1
#define NODE_FOUND     2
#define JUNCTION_FOUND 3

// Global variables for light intensity ranges.

// light intensity for black line
int line;
// light intensity for the ground
int ground;
// light intensity for the junction point
int junction;
// light intensity for the node point
int node;


// buffer to store incoming messages
char buff[256];
int data_length;

// flag to indicate new data
int new_data = 0;
int lock =0;
int dest=5;

// flag to indicate line found
int line_found=0;


typedef struct data_pack comm_pack;

void message_handler(const unsigned char *data, unsigned char length){
	
	// check for length .. ignore message if not of proper length
	if(length != sizeof(comm_pack)){
		cls(); lcd_int(888);
		return;		
	}

	if(new_data == 0 && lock == 0){
		memcpy(buff, data, length);
		data_length = length;
		new_data = 1;
	}

}

/* Motor Functions */
void mot_speed(int speed, int which){
	switch(which){
		case MOT_A: motor_a_speed(speed);
			    break;
		case MOT_C: motor_c_speed(speed);
			    break;
		case ALL  : motor_a_speed(speed);
			    motor_c_speed(speed);
			    break;
	}
}

void turn(int dir){
	switch(dir){
		case LEFT:	motor_a_dir(REV);
				motor_c_dir(FWD);
				break;
		case RIGHT:	motor_a_dir(FWD);
				motor_c_dir(REV);
				break;
		case FRWD:	motor_a_dir(FWD);
				motor_c_dir(FWD);
				break;
		case REVS:	motor_a_dir(REV);
				motor_c_dir(REV);
				break;
	}
}

/* End of Motor Functions */



/* Event definition for various events */

wakeup_t WaitForData(wakeup_t data){
	return new_data;
}

wakeup_t off_the_line(wakeup_t data){
	int isGround,isNode,isJunction;
	isGround = (LIGHT_2 <= ground+1) && (LIGHT_2 >= ground-1);
	isNode = (LIGHT_2 <= node+1) && (LIGHT_2 >= node-1);
	isJunction = (LIGHT_2 <= junction+1) && (LIGHT_2 >= junction-1);
	
	if(isGround)
		return GROUND_FOUND;
	if(isNode)
		return NODE_FOUND;
	if(isJunction)
		return JUNCTION_FOUND;

	return 0; 

}

wakeup_t dark_found(wakeup_t data){
	return ((LIGHT_2 <= line+1) && (LIGHT_2 >= line-1));	
}

/* End of event definitions */

int check_rots(dir, rots){
	if(dir ==  RIGHT)
		return (ROTATION_3 <= rots);
	else
		return (ROTATION_3 >= rots);
}
		


static int wait_dark(int dir, int rots){
	while(check_rots(dir, rots) && !line_found){
		lcd_int(ROTATION_3);
		msleep(1);
		line_found=dark_found(0);		
	}
	return line_found;
}

int getRotVal(int dir, int rots){
	if(dir == LEFT)
		return -rots;
	else
		return rots;
}

void follow_path(comm_pack *data_ptr){
	int mask = 1;
	int node_num=0;
	int retval;
	int turn_dir=LEFT;
	int dir=0;
	int rots=0;
	int junc_found=0;
	dest = data_ptr->b;
	while(node_num != dest){
		turn(FRWD);
		mot_speed(SPEED,ALL);
		line_found = 0;

		retval = wait_event(off_the_line, 0);			
		switch(retval){
			case GROUND_FOUND:
						rots=10;
						dir=LEFT;
						if(junc_found == 1){
							if(turn_dir == LEFT)
								dir = LEFT;
							else if(turn_dir == RIGHT)
								dir = RIGHT;

							rots = 28;
							junc_found=0;
						}
	
						while(!line_found){
							ds_rotation_set(&ROTATION_SENSOR, 0);	
							turn(dir);
							if(wait_dark(dir, getRotVal(dir,rots)))
								break;		 	        // break out of while 

							turn(1-dir);
							if(wait_dark(1-dir, getRotVal(1-dir, 2*rots)))
								break;				// break out of while

							// come back to original position
							turn(dir);
							if(wait_dark(dir, 0))
								break;			// break out of while 
						

							rots +=5; // increase sweep by 5 rotations
	
						} 					
						// on finding line, reset rotation count
						mot_speed(0, ALL);
						ds_rotation_set(&ROTATION_SENSOR, 0);
						break;			

			case NODE_FOUND:
						node_num++;
						if(node_num == dest){
							goto out; 	// can't insert a break for while in switch
						}
						cputs("node");
						break;
				
				
			case JUNCTION_FOUND:
						junc_found =1;
						lcd_int(data_ptr->path);
						sleep(500);cls();
						turn_dir = data_ptr->path & mask;
						
						if(turn_dir == LEFT){
							cputs("left");
						}else{
							cputs("rigt");
						}
						mask = mask << 1;
						break;				


			default: 		lcd_int(888);
						sleep(2);break;

		}
	}
	
out:
	// stop
	mot_speed(0, ALL);


}


int ControllerThread(){
	comm_pack *data_ptr;	

	while(!shutdown_requested()){
		wait_event(WaitForData, 0);		
		lock=1;				// lock the data

		data_ptr = (comm_pack *)&buff;

		follow_path(data_ptr);		
		new_data=0;
	}

	return 0;
}

int CalibrationThread(){
	
	// Take in the readings for line, ground and junction one by one
	
	wait_event(dkey_pressed, KEY_PRGM);
	line = LIGHT_READ;
	
	wait_event(dkey_pressed, KEY_VIEW);
	ground = LIGHT_READ;

	wait_event(dkey_pressed, KEY_PRGM);
	junction = LIGHT_READ;

	wait_event(dkey_pressed, KEY_VIEW);
	node = LIGHT_READ;

	// debug
	lcd_int(line);
	sleep(2);
	lcd_int(ground);
	sleep(2);
	lcd_int(junction);
	sleep(2);
	lcd_int(node);
	sleep(2);
	return 0;
}

int main(int argc, char *argv[]){
	// declare thread ids
	int controller_tid, calibration_tid;

	// Activate light sensor
	ds_active(&LIGHT_SENSOR);
	
	// Activate rotational sensor
	ds_active(&ROTATION_SENSOR);
	ds_rotation_on(&ROTATION_SENSOR);
	ds_rotation_set(&ROTATION_SENSOR, 0);

	// set logical range of ir comm
	lnp_logical_range(NEAR);

	// register ir message handler
	lnp_integrity_set_handler(message_handler);

	controller_tid  = execi(&ControllerThread, 0, NULL, PRIO_NORMAL, DEFAULT_STACK_SIZE);
	calibration_tid = execi(&CalibrationThread, 0, NULL, PRIO_NORMAL, DEFAULT_STACK_SIZE);

	while(!shutdown_requested()){}

	// Kill controller thread. Calibration thread will exit itself
	kill(controller_tid);

}



#endif

